
import numpy as np
import os
import glob
import yaml
import pandas as pd
import xarray as xr

def FilePattern(var, levtype, freq, yaml_content, rundir, start_date, end_date):
   start_date_str = str(start_date)
   end_date_str = str(end_date)
   model = yaml_content['model']
   project = yaml_content['project']
   simulation = yaml_content['simulation']
   ensemble = yaml_content['ensemble']
   
   file_pattern = levtype + "_*" + freq + "*" + project + "_" + model + "_" + simulation + "_" + ensemble + "_" + var + "_" + start_date_str + "-" + end_date_str
   file_pattern = rundir + "*" + file_pattern


   return file_pattern

def Construct_Dates(start_date, end_date, frequency, rundir):
    # DATE: YYYYMMDDZZ # 
    start_date = str(start_date)
    end_date = str(end_date)
    start_date_date = start_date[:8]
    start_date_time = start_date[8:]
    start_date_reformat = f"{start_date_date[:4]}-{start_date_date[4:6]}-{start_date_date[6:]} {int(start_date_time):02d}:00:00"

    end_date_date = end_date[:8]
    end_date_time = end_date[8:]
    end_date_reformat = f"{end_date_date[:4]}-{end_date_date[4:6]}-{end_date_date[6:]} {int(end_date_time):02d}:00:00"

    rundir = rundir + "/02-InterpolatedGrid"
    date_array = []
    if frequency == "DAILY":
        date_array = pd.date_range(start=start_date_reformat, end=pd.to_datetime(end_date_reformat) - pd.Timedelta(days=1), freq='D')
        date_array = date_array.strftime('%Y%m%d%H')
        date_array = date_array.to_numpy()
        rundir = rundir + "/03-Daily/"

    elif frequency == "6HOURLY":  # CAREFUL, WHEN REACHING 24 IT GOES TO 30
        start_date_time = "06"    # Start date of hourly output is not 00, but 06 for some reason
        start_date_reformat = f"{start_date_date[:4]}-{start_date_date[4:6]}-{start_date_date[6:]} {int(start_date_time):02d}:00:00"

        end_date_date = end_date[:8]
        end_date_time = end_date[8:]
        end_date_reformat = f"{end_date_date[:4]}-{end_date_date[4:6]}-{end_date_date[6:]} {int(end_date_time):02d}:00:00"
        date_array = pd.date_range(start=start_date_reformat, end=end_date_reformat, freq='6h',)
        date_array = date_array.strftime('%Y%m%d%H')
        date_array = date_array.to_numpy()
        rundir = rundir + "/02-6Hourly/"

    elif frequency == "MONTHLY":
        date_array = pd.date_range(start=start_date_reformat, end=end_date_reformat, freq='MS')
        date_array = date_array.strftime('%Y%m%d%H')
        date_array = date_array.to_numpy()
        rundir = rundir + "/04-Monthly/"
    else:
        print("ERROR, no correct frequency for date reconstruction")
        return []
    return date_array, rundir

### Refactoring code to improve readability and logic ###
def check_file(pattern):
    #print(f"Searching for pattern '{pattern}'")
    matching_files = glob.glob(pattern)
    
    
    if matching_files:
        found = matching_files
        missing = ""
#        print(f"Found files with pattern '{pattern}'")
    else:
        missing = pattern
        found = ""
        print(f"No files matched the pattern '{pattern}'")

    return found, missing

def process_level(level, date_pattern, level_dim_name):
    final_pattern = f"{date_pattern}_level_{level}.nc"
    found, missing = check_file(final_pattern)
    if missing:
        return None, missing
    ds = xr.open_mfdataset(found[0], engine='netcdf4', mask_and_scale=False, decode_times=False)
    ds = ds.assign_coords(level_dim_name = level).expand_dims(dim=level_dim_name)        # Could be different depending on if PL, or O3d
    return ds, None

def process_level_fields(lev, levels, date_pattern, level_dim_name):

    ds_levels =  []
    missing_files = []

    for level in levels:
        ds_level, missing = process_level(level, date_pattern, level_dim_name)
        if missing:
            missing_files.append(missing)
        elif ds_level is not None:
            ds_levels.append(ds_level)
    
    if ds_levels:
        print(f"Concatenating all levels for date")
        ds_combined_levels = xr.concat(ds_levels, dim=level_dim_name) # generalize for ocean
        if level_dim_name == 'plev19':
            ds_combined_levels['plev19'].attrs.update({
                'units': 'Pa',
                'long_name': 'pressure',
                'standard_name': 'air_pressure',
                'positive': 'down'
            })
        elif level_dim_name == 'olevel':
            NEMO_levels_file = "/gpfs/scratch/bsc32/bsc032519/EERIE_tools/post-processing-netcdf-files/NEMO_levels_m.nc"   ### TODO: change the path, it is inside the git repo
            NEMO_levels = xr.open_dataset(NEMO_levels_file)
            list_of_levels = NEMO_levels["nav_lev"].values
            list_of_levels = dict(enumerate(list_of_levels, 1))
            ds_combined_levels['olevel'] = ('olevel', [list_of_levels[level]])
            ds_combined_levels['olevel'].attrs.update({
                'units': 'm',
                'long_name': 'depth',
                'standard_name': 'depth',
                'positive': 'down'
            })
        return ds_combined_levels, missing_files
    return None, missing_files

def process_single_level_fields(lev, date_pattern):
    final_pattern = date_pattern + "*.nc" # maybe * is not needed
    found, missing = check_file(final_pattern)
    if missing:
        print(f"Some files are missing for variable")
        return None, missing
    else:
        print(f"Found file is '{found[0]}'")
        ds = xr.open_mfdataset(found[0], engine='netcdf4', mask_and_scale=False, decode_times=False)
        return ds, None
    



def process_date(date, file_pattern, levtype):
    lev = levtype.get('levtype')
    date_pattern = f"{file_pattern-{date}}"
    lev_bool = False
    if lev == "PL":
        levels = np.array([1000, 925, 850, 700, 600, 500, 400, 300, 250, 200, 150, 100, 70, 50, 30, 20, 10, 5, 1])
        levels = levels * 1000             # To get Pa instead of hPa
        level_dim_name = 'plev19'
        ds_combined, missing_files = process_level_fields(lev, levels, date_pattern, level_dim_name)
    elif lev == "OCEAN":
        level_dim_name = 'olevel'
        n_levels = levtype.get('3d_levels')
        if n_levels == "lev5":
            levels = np.array([18, 30, 41, 46, 50])
            print("Going for reduced levels\n")
        else:
            levels = np.arange(1, 76)
        ds_combined, missing_files = process_level_fields(lev, levels, date_pattern, level_dim_name)
    else:
        ds_combined, missing_files = process_single_level_fields(lev, date_pattern)   # Just returns a single netcdf file, does not open it


    return ds_combined, missing_files

def process_variable(var, levtype, freq, content, rundir):
    lev = levtype.get('levtype')
    print(f"Levtype is '{lev}'")
    start_date = yaml_content['start_date']       ## Careful when we are getting this info
    end_date = yaml_content['end_date']
    dates, rundir = Construct_Dates(start_date, end_date, freq, rundir)
    print(f"Rundir is '{rundir}'")
    file_pattern = FilePattern(var, lev, freq, yaml_content, rundir, start_date, end_date)
    d_req_missing = []
    d_req_found = []
    for date in dates:
        ds, missing = process_date(date, file_pattern, levtype)
        d_req_found.append(ds)
       
    print("Calling xarray to concatenate files")
    print(f"Concatenating files '{d_req_found}'")
    dates_combined = xr.concat(d_req_found, dim="time")
    dates_combined['latitude'].attrs['standard_name'] = 'latitude'
    dates_combined['longitude'].attrs['standard_name'] = 'longitude'
    dates_combined['time'].attrs['standard_name'] = 'time'

    dates_combined = dates_combined.transpose('time', ...)
    name = f"Combined_dates_var_{var}_{start_date}-{end_date}.nc"
    dates_combined.to_netcdf(name)
    

def process_levtype(levtype, freq, content, rundir):


    # Process variables in each levtype

    for var in levtype.get('netcdf_name', []):
        print(f"Variable is '{var}'")
        process_variable(var, levtype, freq, content, rundir)


def process_frequency(freq, content, rundir):
    print(f"Frequency is '{freq}'")
    levtypes = content.get(freq, [])
    # Process each levtype

    for levtype in levtypes:
        process_levtype(levtype, freq, content, rundir)

if __name__ == "__main__":
    data_portfolio = "/gpfs/scratch/bsc32/bsc032519/EERIE_tools/post-processing-netcdf-files/Data-Portfolio-EERIE.yaml"
    rundir =  "/gpfs/scratch/bsc32/bsc032519/RUNDIR/IFS-NEMO/tco79l137/hz9m/hres/intel.undef/mn5-gpp.intel.sp/h768.N4.T96xt4xh1+ioT8xt4xh0.eerie_6h.i32r0w32.eORCA1_Z75.htco79-10708258"
    with open(data_portfolio, 'r') as file:
        yaml_content = yaml.safe_load(file)

    content = yaml_content.get("test", {})
    for freq in content:
        process_frequency(freq, content, rundir)
